import pytest
import numpy as np
import pandas as pd
from numpy.testing import assert_equal, assert_allclose
from sciparse import sampling_period, title_to_quantity, \
         to_standard_quantity, frequency_bin_size, quantity_to_title, \
         dict_to_string, string_to_dict, is_scalar, column_from_unit, \
         cname_from_unit, assertDataDictEqual
from sciparse import assert_allclose_qt, assert_equal_qt, ureg

def test_sampling_period():
    data = pd.DataFrame({'Time (ms)': [0, 0.1, 0.2, 0.3, 0.4],
                         'Values': [0, 1, 2, 3, 4]})
    actual_period = sampling_period(data)
    desired_period = ureg.ms * 0.1
    assert actual_period == desired_period

def test_sampling_period_error():
    data = pd.DataFrame({'Time (ms)': [0],
                         'Values': [0]})
    with pytest.raises(ValueError):
        actual_period = sampling_period(data)

def test_frequency_bin_error():
    data = pd.DataFrame({'frequency (Hz)': [0],
                         'Values': [0]})
    with pytest.raises(ValueError):
        actual_bin = frequency_bin_size(data)

def test_quantity_to_title():
    quantity = ureg.mV*1.0
    desired_title = 'voltage (mV)'
    actual_title = quantity_to_title(quantity)
    assert_equal(actual_title, desired_title)

    quantity = ureg.nA**2*1.0
    desired_title = 'power (nA ** 2)'
    actual_title = quantity_to_title(quantity)
    assert_equal(actual_title, desired_title)

def test_quantity_to_title_with_name():
    quantity = ureg.mV * 1.0
    desired_name = 'photovoltage'
    desired_title = 'photovoltage (mV)'
    actual_title = quantity_to_title(quantity, desired_name)
    assert_equal(actual_title, desired_title)

def testExtractTimeUnits():
    unit_string = 'time (s)'
    desired_unit = 1 * ureg.s
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'time (ms)'
    desired_unit = 1 * ureg.ms
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'time (us)'
    desired_unit = 1 * ureg.us
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'time (ns)'
    desired_unit = 1 * ureg.ns
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'time (ps)'
    desired_unit = 1 * ureg.ps
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

def testExtractElectricalUnits(ureg):
    unit_string = 'Photocurrent (pA)'
    desired_unit = 1 * ureg.pA
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Photocurrent (nA)'
    desired_unit = 1 * ureg.nA
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'current (uA)'
    desired_unit = 1 * ureg.uA
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Jordans (mA)'
    desired_unit = 1 * ureg.mA
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'More Current (A)'
    desired_unit = 1 * ureg.A
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Photovoltage (V)'
    desired_unit = 1 * ureg.V
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Photovoltage (mV)'
    desired_unit = 1 * ureg.mV
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Photovoltage (uV)'
    desired_unit = 1 * ureg.uV
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

    unit_string = 'Photovoltage (nV)'
    desired_unit = 1 * ureg.nV
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

def testExtractSquaredUnits():
    unit_string = 'voltage (mV^2)'
    desired_unit = 1 * ureg.mV ** 2
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

def test_title_to_quantity_squared_2():
    unit_string = 'voltage (mV ** 2)'
    desired_unit = 1 * ureg.mV ** 2
    actual_unit = title_to_quantity(unit_string)
    assert desired_unit == actual_unit

def test_title_to_quantity_name():
    unit_string = 'photovoltage (kV)'
    desired_name = 'photovoltage'
    desired_quantity = 1 * ureg.kV
    actual_quantity, actual_name = title_to_quantity(
            unit_string, return_name=True)
    assert_equal_qt(actual_quantity, desired_quantity)
    assert_equal(actual_name, desired_name)

def testToStandardUnit():
    quantity = 0.1 * ureg.mV
    desired_quantity = 0.1 * 1e-3 * ureg.V
    actual_quantity = to_standard_quantity(quantity)
    assert desired_quantity == actual_quantity

def test_to_standard_quantity_squared():
    quantity = 0.1 * ureg.mV ** 2
    desired_quantity = 0.1 * 1e-6 * ureg.V ** 2
    actual_quantity = to_standard_quantity(quantity)
    assert desired_quantity == actual_quantity

def test_to_standard_quantity_psd():
    quantity = 0.1 * ureg.mA ** 2 / ureg.Hz
    desired_quantity = 0.1 * 1e-6 * ureg.A ** 2 / ureg.Hz
    actual_quantity = to_standard_quantity(quantity)
    assert desired_quantity == actual_quantity

def test_frequency_bin_size():
    psd_data = pd.DataFrame({
            'frequency (Hz)': [1.5, 3.0, 4.5],
            'power (V^2)': [0, 1, 2]})
    actual_quantity = frequency_bin_size(psd_data)
    desired_quantity = 1*ureg.Hz*1.5
    assert actual_quantity == desired_quantity

def test_dict_to_string():
    metadata = {
        'wavelength': 10 * ureg.nm,
        'material': 'Al',
        'replicate': 2}

    actual_string = dict_to_string(metadata)
    desired_dict = {
        'wavelength (nm)': 10,
        'material': 'Al',
        'replicate': 2}
    assert_equal(actual_string, str(desired_dict))

def test_string_to_dict():
    input_string = "{'wavelength (nm)': 10, 'material': 'Al', 'replicate': 2}"
    desired_dict = {
        'wavelength': 10*ureg.nm,
        'material': 'Al',
        'replicate': 2}
    actual_dict = string_to_dict(input_string)
    assertDataDictEqual(actual_dict, desired_dict)

def test_is_scalar():
    quantity = 5 * ureg.Hz
    data_scalar_actual = is_scalar(quantity)
    data_scalar_desired = True
    assert_equal(data_scalar_actual, data_scalar_desired)

    quantity = 6.0 * ureg.Hz
    data_scalar_actual = is_scalar(quantity)
    data_scalar_desired = True
    assert_equal(data_scalar_actual, data_scalar_desired)

    quantity = 6.0
    data_scalar_actual = is_scalar(quantity)
    data_scalar_desired = True
    assert_equal(data_scalar_actual, data_scalar_desired)

    quantity = np.array([6.0, 2.5]) * ureg.Hz
    data_scalar_actual = is_scalar(quantity)
    data_scalar_desired = False
    assert_equal(data_scalar_actual, data_scalar_desired)

def test_is_scalar_complex():

    quantity = (1 + 2j) * ureg.ohm
    data_scalar_actual = is_scalar(quantity)
    data_scalar_desired = True
    assert_equal(data_scalar_actual, data_scalar_desired)

def test_column_from_unit():
    input_data = pd.DataFrame({
        'Time (ms)': [0, 1, 2, 3],
        'Photovoltage (nV)': [0, 1, 4, 5]})
    desired_data = ureg.uV * 1e-3 * np.array([0., 1, 4, 5])
    actual_data = column_from_unit(input_data, ureg.uV)
    assert_allclose_qt(actual_data, desired_data, atol=1e-12)

    desired_data = ureg.s * 1e-3 * np.array([0, 1, 2, 3])
    actual_data = column_from_unit(input_data, ureg.s)
    assert_allclose_qt(actual_data, desired_data, atol=1e-12)

def test_column_from_unit_extra():
    input_data = pd.DataFrame({
        'Time': [0, 1, 2, 3],
        'Photovoltage (nV)': [0, 1, 4, 5],
        'Sync': [0, 0, 2, 1]
        })
    desired_data = ureg.uV * 1e-3 * np.array([0., 1, 4, 5])
    actual_data = column_from_unit(input_data, ureg.uV)
    assert_allclose_qt(actual_data, desired_data, atol=1e-12)

def test_column_not_found():
    input_data = pd.DataFrame({})
    with pytest.raises(ValueError):
        column_from_unit(input_data, ureg.ms)

def test_cname_from_unit():
    input_data = pd.DataFrame({
        'Time': [0, 1, 2, 3],
        'Photovoltage (nV)': [0, 1, 4, 5],
        'Sync': [0, 0, 2, 1]
        })
    desired_name = 'Photovoltage (nV)'
    actual_name = cname_from_unit(input_data, ureg.V)
    assert_equal(actual_name, desired_name)



import re
import pandas as pd
import numpy as np
from sciparse import string_to_dict, dict_to_string, ureg
from datetime import datetime as dt
import pint

def parse_default(
        filename, data=None, metadata=None, read_write='r'):
    """
    Default parser for data. Assumes there is a single line of metadata at the top line in the form of a dictionary and the remainder is tabular and can be imported as a pandas DataFrame

    :param filename: Name of the file to be written
    :param data: Data to write to file
    :param metadata: Metadata to write to file
    :param read_write: "r" or "w". Read or write.
    """

    if read_write == 'r':
        with open(filename) as fh:
            string_metadata = fh.readline().rstrip('\n')
            metadata = string_to_dict(string_metadata)
            data = pd.read_csv(fh)

        return data, metadata
    elif read_write == 'w':
        with open(filename, 'w') as fh:
            metadata_line = dict_to_string(metadata) + '\n'
            fh.write(metadata_line)
            if isinstance(data, pd.DataFrame):
                data.to_csv(fh, mode='a', index=False)
            elif data is None:
                pass
            else:
                raise ValueError(f'This method only implemented for type of pd.DataFrame, you attempted to pass in type {type(data)}.')

def parse_xrd(filename):
    """
    Loads XRD data from a given filename

    :param filename: The filename to open
    :returns (data, metadata): set of data and metadata from XRD file

    """
    with open(filename) as file_handle:
        count = 0
        reg_pattern = re.compile(r'\w+=\S*\w+')
        metadata = {}
        while True:
            line = file_handle.readline()
            if line == '[Data]\n' or not line:
                break
            line = line.rstrip('\n')

            if reg_pattern.match(line):
                split_string = line.split('=')
                name = split_string[0].lower()
                val = split_string[1].lower()
                interesting_names = [
                    'drivename', 'startposition', 'date',
                     'time', 'increment', 'scantype','start',
                     'steps']
                if name in interesting_names:
                    if name == 'drivename':
                        current_drive = val
                    elif name == 'startposition':
                        metadata[current_drive] = float(val)
                    else:
                        try:
                            metadata[name] = int(val)
                        except:
                            try:
                                metadata[name] = float(val)
                            except:
                                metadata[name] = val
        data = pd.read_csv(file_handle)
        data = data.loc[:, ~data.columns.str.contains('^Unnamed')]
        data = data.rename(columns=lambda x: x.strip()) # Strip whitespace
        data = data.rename(columns={'Angle': 'Angle (deg)', 'Det1Disc1':
                                    'Counts'})
        return data, metadata

def parse_lcr(filename):
    metadata = parse_lcr_header(filename)
    data_start_line = find_lcr_dataline(filename)

    with open(filename, 'r') as dataFile:
        lines = dataFile.readlines()
        header_line = lines[data_start_line-1]
        header_line = header_line.replace("\n", "")
        data_types = header_line.split("\t")
        data_types = [x for x in data_types if x != ""]
        data = pd.DataFrame({})
        for dtype in data_types:
            data[dtype] = np.array([])

        lines = lines[data_start_line:-4]
        if len(lines) != metadata['n_samples']:
            raise ValueError("number lines not equal to data lines. Issue with parsing.")

        for line in lines:
            line = line.replace("\n", "")
            line_data = line.split("\t")
            line_data = [float(x) for x in line_data if x != '']
            new_data_row = pd.DataFrame({})
            for i, dtype in enumerate(data_types):
                new_data_row[dtype] = [line_data[i]]
            data = data.append(new_data_row)

    return data, metadata

def parse_lcr_header(filename):
    metadata = {}
    currentID = ''
    with open(filename, 'r') as dataFile:
        lines = dataFile.readlines()
        for line in lines:
            line = line.replace(":", " ")
            line = line.replace("\n", "")
            line = line.replace("\t", "")
            lineData = line.split(" ")
            if lineData[0] == 'ID':
                currentID = lineData[-1]
            if lineData[0] == 'MODE':
                if currentID == 'A':
                    mode = lineData[-1]
            elif lineData[0] == 'START':
                start_voltage = float(lineData[-1])
            elif lineData[0] == 'STOP':
                stop_voltage = float(lineData[-1])
            elif lineData[0] == 'STEP':
                step_voltage = float(lineData[-1])
            elif lineData[0] == 'DATE':
                recorded_date = dt.strptime(lineData[-1], "%m/%d/%Y")
            elif lineData[0] == 'TIME':
                timeData = lineData[-3] + " " + lineData[-2] + " " + lineData[-1]
                recorded_time = dt.strptime(timeData, "%H %M %S")
            elif lineData[0] == 'PNTS':
                n_samples = int(lineData[-1])
    metadata = {
        'start_voltage': start_voltage,
        'stop_voltage': stop_voltage,
        'step_voltage': step_voltage,
        'date': recorded_date,
        'time': recorded_time,
        'n_samples': n_samples,
        'mode': mode,
    }
    return metadata

def find_lcr_dataline(filename):
    with open(filename, 'r') as dataFile:
        lines = dataFile.readlines()
        currentLine = 0
        dataLines = []
        for line in lines:
            lineData = line.split(" ")
            if lineData[0] == "DATA:": # begin reading data
                data_start_line = currentLine + 3
                return data_start_line
            currentLine += 1

def convert_lcr_to_standard(data, metadata):
    """
    Converts an LCR dataset into standard form (Z, theta).

    :param data: Pandas DataFrame containing LCR data in a potentially non-standard form (Cp/Q, Cp/D, Cp/Rp for example)
    """
    columns = data.columns.values
    if 'frequency' not in metadata and 'f' not in metadata:
        raise ValueError(f'Frequency not in metadata. Metadata contains {metadata}')
    f = metadata['frequency']
    if isinstance(f, pint.Quantity):
        f = f.to(ureg.Hz).m
    w = 2*np.pi*f
    new_data = data.copy()

    if 'Z' in columns and 'THETA' in columns:
        Z = data['Z']
        theta = data['THETA']*np.pi / 180
        del new_data['Z'], new_data['THETA']
    elif 'C' in columns and 'Q' in columns:# C assumed to be Cp
        Q = new_data['Q']
        Cp = new_data['C']
        Z = 1 / (w*Cp) * 1 / np.sqrt(1 + 1/(Q**2))
        theta = -np.arctan(Q)
        del new_data['Q'], new_data['C']
    elif 'C' in columns and 'D' in columns: # D = 1/Q. C assumed Cp
        D = data['D']
        Cp = data['C']
        Z = 1 / (w*Cp) * 1 / np.sqrt(1 + (D**2))
        theta = -np.arctan(1 / D)
        del new_data['D'], new_data['C']
    elif 'CP' in columns and 'RP' in columns:
        Rp = data['RP']
        Cp = data['CP']
        Z = 1 / np.sqrt((w*Cp)**2 + 1 / Rp**2)
        theta = -np.arctan(w*Rp*Cp)
        del new_data['RP'], new_data['CP']
    elif 'CS' in columns and 'RS' in columns:
        Rs = data['RS']
        Cs = data['CS']
        Z = np.sqrt(Rs**2 + 1/((w*Cs)**2))
        theta = -np.arctan(1 / (w*Rs*Cs))
        del new_data['RS'], new_data['CS']
    else:
        raise ValueError(f'Columns did not match any available impedance models. Available models are CP/Q, CP/D, CP/RP, CS/RS. Column of data are {columns}')

    new_data['Z (ohm)'] = Z
    new_data['THETA (rad)'] = theta
    new_data = new_data.rename(
            columns={
            'VM': 'VM (V)',
            'IM': 'IM (A)',
            'BIAS': 'BIAS (V)',
            })
    return new_data



import pint
from numpy.testing import assert_equal, assert_allclose
import numpy as np
from pandas.testing import assert_frame_equal
from pandas import DataFrame
import warnings
from itertools import zip_longest

def assert_equal_qt(actual_data, desired_data, atol=1e-6, rtol=1e-6):
    if isinstance(actual_data, np.ndarray):
        assert_equal(actual_data, desired_data)
    elif isinstance(actual_data, pint.Quantity) or \
            isinstance(desired_data, pint.Quantity):
        assert_equal(isinstance(actual_data, pint.Quantity), True)
        assert_equal(isinstance(desired_data, pint.Quantity), True)
        assert_equal(actual_data.to(desired_data.units).magnitude, \
                desired_data.magnitude)
        assert actual_data.to(desired_data.units).units == desired_data.units

def assert_allclose_qt(
        actual_data, desired_data, atol=1e-15, rtol=1e-14):
    if isinstance(actual_data, np.ndarray):
        assert_allclose(
                actual_data, desired_data, atol=atol, rtol=rtol)
    elif isinstance(actual_data, pint.Quantity) or \
            isinstance(desired_data, pint.Quantity):
        assert_equal(isinstance(actual_data, pint.Quantity), True)
        assert_equal(isinstance(desired_data, pint.Quantity), True)
        assert_allclose(
                actual_data.magnitude, desired_data.magnitude,
                atol=atol, rtol=rtol)
        assert actual_data.units == desired_data.units

def assertDataDictEqual(data_dict_actual, data_dict_desired):
# Check that they have the exact same set of keys
    assert_equal(type(data_dict_actual), dict)
    assert_equal(type(data_dict_desired), dict)
    actual_keys = set(data_dict_actual.keys())
    desired_keys = set(data_dict_desired.keys())
    assert_equal(actual_keys, desired_keys)

    for actual_key, actual_val in data_dict_actual.items():
        desired_val = data_dict_desired[actual_key]
        if isinstance(actual_val, DataFrame):
            assert_frame_equal(actual_val, desired_val)
        elif isinstance(actual_val, dict):
            assertDataDictEqual(actual_val, desired_val)
        else:
            assert_equal_qt(actual_val, desired_val)



import pytest
import pandas as pd
from pandas.testing import assert_frame_equal
from numpy.testing import assert_equal
from sciparse import parse_xrd, parse_default, ureg
import os

@pytest.fixture
def test_dir():
    file_location = os.path.dirname(os.path.abspath(__file__))
    location = file_location + '/data/'
    yield location
    filename = location + 'TEST1~wavelengths-1~temperatures-25.csv'
    try:
        os.remove(filename)
    except:
        print("default file not found. That's OK, we're creating it on the fly")

def test_load_xrd_data(test_dir):
    xrd_filename = test_dir + 'TEST1~1~type-locked_coupled~peak-Si.txt'
    data_desired = pd.DataFrame({
        'Angle (deg)': [69.05, 69.055, 69.06, 69.065, 69.07,69.075,69.08,
        69.085, 69.09, 69.095, 69.1, 69.105, 69.11, 69.115],
        'Counts': [24, 30, 28, 40, 132, 272, 3472, 16368,21970,10562,
                   1210,264,130,64]})
    data_actual, _ = parse_xrd(xrd_filename)
    assert_frame_equal(data_actual, data_desired)

def test_load_xrd_metadata(test_dir):
    xrd_filename = test_dir + 'TEST1~1~type-locked_coupled~peak-Si.txt'
    metadata_desired = {
        'date': '02/10/2021',
        'increment': 0.005, 'scantype': 'locked coupled',
        'start': 69.05, 'steps': 14, 'time': 1,
        'theta': 34.0, '2theta': 68.0, 'phi': 180.13, 'chi': -0.972}
    _, metadata_actual = parse_xrd(xrd_filename)
    assert_equal(metadata_actual, metadata_desired)

def test_load_metadata_default(test_dir):
    """
    Tests that we can load metadata from a file successfully
    """
    filename = test_dir + 'TEST1~wavelengths-1~temperatures-25.csv'
    with open(filename, 'w+') as fh:
        fh.write('{"frequency": 8500}\n')
        fh.write(f'Time, Data\n')
        fh.write(f'1, 2\n')

    _, metadata_actual = parse_default(filename)
    metadata_desired = {'frequency': 8500}
    assert_equal(metadata_actual, metadata_desired)

def test_load_data(test_dir):
    data_desired = pd.DataFrame({'wavelengths': [1, 2, 3],
                             'Mean': [1,2,4]})
    filename = test_dir + 'TEST1~wavelengths-1~temperatures-25.csv'
    metadata_desired = {'frequency': 8500}
    with open(filename, 'w+') as fh:
        fh.write(str(metadata_desired) + '\n')
        data_desired.to_csv(fh, mode='a', index=False)

    data_actual, _ = parse_default(filename)
    assert_frame_equal(data_actual, data_desired)

def test_parse_default_save_metadata(test_dir):
    metadata = {'wavelength': 10*ureg.nm, 'current': 100*ureg.uA}
    filename = test_dir + 'TEST1~wavelengths-1~temperatures-25.csv'
    parse_default(filename, metadata=metadata, read_write='w')
    with open(filename, 'r') as fh:
        metadata_actual= fh.readline()
    metadata_desired = "{'wavelength (nm)': 10, 'current (µA)': 100}\n"
    assert_equal(metadata_actual, metadata_desired)

def test_parse_default_save_data(test_dir):
    data_desired = pd.DataFrame({'wavelengths': [1, 2, 3],
                             'Mean': [1,2,4]})
    metadata = {'wavelength': 10*ureg.nm, 'current': 100*ureg.uA}
    filename = test_dir + 'TEST1~wavelengths-1~temperatures-25.csv'
    parse_default(filename, data=data_desired, metadata=metadata, read_write='w')
    with open(filename, 'r') as fh:
        metadata = fh.readline()
        data_actual = pd.read_csv(fh)
    assert_frame_equal(data_actual, data_desired)



import pint
import re
from ast import literal_eval
from sciparse import ureg
import numpy as np

def sampling_period(data):
    """
    Extracts the sample period from a pandas DataFrame assuming that data is evenly spaced in time. Uses the difference in time between the first and second data points to determine the sampling period  and sampling frequency.

    :param data: Pandas array with Time (including units, i.e. Time (ms))
    :returns sampling_period: Sampling period as a pint quantity
    """
    # For now, assume time is the first variable in the column
    columns_names = data.columns.values
    time_column = 0
    time_string = columns_names[time_column]
    if len(data) < 2:
        raise ValueError(f'Data does not have enough points to compute a sampling period. Only has {len(data)} ponits')
    delta_time = data[time_string][1] - data[time_string][0]

    sampling_period = delta_time * title_to_quantity(time_string)
    return sampling_period

def frequency_bin_size(data):
    """
    Extracts the frequency bin (in Hz, as a pint unit) from a PSD dataset.

    :param data: Input power spectral density as a Pandas Array
    :returns bin_size: frequency bin size as a pint unit
    """
    column_names = data.columns.values
    frequency_column = 0
    frequency_string = column_names[frequency_column]
    if len(data) < 2:
        raise ValueError(f'Data does not have enough points to compute a frequency bin size. Only has {len(data)} ponits')
    delta_frequency = data[frequency_string].iloc[1] - data[frequency_string].iloc[0]

    bin_size = delta_frequency * title_to_quantity(frequency_string)
    return bin_size

def title_to_quantity(title_string, return_name=False):
    """
    Extracts unit from a column name such as Time (ms).

    :param title_string: The string to look for the unit
    :param return_name: Whether to return a tuple containing the base name and the unit.
    """
    # Search for anything in parentheses, interpret as a unit.
    # Squared units are fine.
    unit_pattern = re.compile(r'\(.*\)') # Find the thing in parentheses.
    match = unit_pattern.search(title_string)
    if match == None:
        quantity = ureg.Quantity(1)
        bare_unit_string = ''
        unit_string_parens = ''
    else:
        bare_unit_string = match.group()[1:-1] # remove parentheses
        unit_string_parens = match.group()
        quantity = 1 * ureg.parse_expression(bare_unit_string)

    if return_name:
        name = title_string.replace(unit_string_parens , '').rstrip()
        return quantity, name
    else:
        return quantity

def quantity_to_title(quantity, name=None):
    """
    Converts a quantity into a standard title
    """
    standard_mapping = {
        ureg.V: 'voltage',
        ureg.A: 'current',
        ureg.ohm: 'impedance',
        ureg.V**2: 'power',
        ureg.A**2: 'power',
        ureg.V**2/ureg.Hz: 'PSD',
        ureg.A**2/ureg.Hz: 'PSD',
        ureg.ohm**2: 'power',
        ureg.Hz: 'frequency',
        ureg.m: 'wavelength',
        ureg.s: 'time',
        ureg.degK: 'temperature',
        ureg.degC: 'temperature',
        ureg.degF: 'temperature',
        ureg.J: 'energy',
        ureg.J / ureg.m ** 3: 'energy density',
        ureg.J / ureg.m ** 2: 'energy density',
        ureg.J / ureg.m: 'energy density',
        ureg.W: 'power',
        ureg.W / ureg.m ** 3: 'power density',
        ureg.W / ureg.m ** 2: 'power density',
        ureg.W / ureg.m: 'power density',
        ureg.J / ureg.K: 'entropy',
        ureg.N: 'force',
        ureg.N / ureg.m ** 2: 'pressure',
        ureg.Pa: 'pressure',
        ureg.m / ureg.s **2: 'acceleration',
        ureg.m / ureg.s: 'velocity',
        ureg.m ** 2: 'area',
        ureg.m ** 3: 'volume',
        ureg.kg / ureg.m ** 3: 'density',
        ureg.C: 'charge',
        ureg.C / ureg.m ** 3: 'charge density',
        ureg.C / ureg.m ** 2: 'charge density',
        ureg.C / ureg.m: 'charge density',
        ureg.A / ureg.m ** 2: 'current density',
        ureg.A / ureg.m: 'current density',
        ureg.V / ureg.m: 'field',
        ureg.A / ureg.m: 'field',
        ureg.T: 'field',
        ureg.Wb: 'flux',
        ureg.ohm: 'resistance',
        ureg.ohm * ureg.m: 'resistivity',
        ureg.siemens: 'conductance',
        ureg.siemens / ureg.m: 'conductivity',
        ureg.H: 'inductance',
        ureg.F: 'capacitance',
        ureg.K / ureg.W: 'thermal conductivity',
        ureg.Unit('dimensionless'): '',
    }
    if not name:
        standard_unit = to_standard_quantity(quantity).units
        title_prefix = standard_mapping[standard_unit]
    else:
        title_prefix = name

    if quantity.dimensionless:
        title = title_prefix
    else:
        title = title_prefix + ' ({:~})'.format(quantity.units)

    return title

def to_standard_quantity(quantity):
    """
    :param quantity: Pint quantity in non-standard form (i.e. mV)
    """
    unit = quantity.units
    if quantity.dimensionless:
        base_unit = ureg.Quantity(1)

    unit_str_list = re.split(r' / | \* ', str(unit))
    quantity_list = [ureg.parse_expression(unit_str) \
            for unit_str in unit_str_list]

    if len(quantity_list) == 1 and not quantity.dimensionless:
        target_quantity = quantity_list[0]
        contains_power = '**' in str(target_quantity)
        if contains_power: # We have a power to deal with
            unit_power = int(str(unit)[-1])
            unit_base_str = str(unit)[:-5]
            return_tuple = ureg.parse_unit_name(str(unit_base_str))
            (prefix, base_unit, _) = return_tuple[0]
            base_unit = ureg.Quantity(1, base_unit)**unit_power
        else:
            return_tuple = ureg.parse_unit_name(str(unit))
            (prefix, base_unit, _) = return_tuple[0]

    elif len(quantity_list) > 1 and not quantity.dimensionless:
        standard_quantity_list = \
            [to_standard_quantity(q) for q in quantity_list]
        multiplication_locations = \
            [x.span()[0] for x in re.finditer(r' \* ', str(quantity))]
        division_locations = \
            [x.span()[0] for x in re.finditer(r' / ', str(quantity))]
        operator_locations = multiplication_locations + division_locations
        operator_locations.sort()
        base_unit = standard_quantity_list[0]
        for q in standard_quantity_list[1:]:
            next_index = operator_locations.pop(0)
            if next_index in multiplication_locations:
                base_unit *= q
            elif next_index in division_locations:
                base_unit /= q


    new_quantity = quantity.magnitude * ureg.Quantity(1, unit).to(base_unit)
    return new_quantity

def is_scalar(quantity):
    data_is_scalar =  isinstance(quantity,
            (int, float, complex, np.float, np.integer, np.complex))
    if isinstance(quantity, pint.Quantity):
        data_is_scalar = isinstance(quantity.magnitude, (int, float, complex))
    return data_is_scalar

def dict_to_string(metadata):
    """
    Converts a dictionary, potentially containing pint units, into a human- and machine-readable string

    :param metadata: Dictionary of metadata to be saved
    """
    dict_to_write = {}
    for k, v in metadata.items():
        if isinstance(v, pint.Quantity):
            name_string = quantity_to_title(v, name=k)
            value = v.magnitude
        elif isinstance(v, pint.Unit):
            name_string = k
            value = str(v)
        else:
            name_string = str(k)
            value = v
        dict_to_write[name_string] = value
    return str(dict_to_write)

def string_to_dict(metadata_string):
    """
    Converts a human-readable string into a dictionary, with pint units as applicable.

    :param metadata_string: The string to be converted into a set of metadata
    """
    input_dict = literal_eval(metadata_string)
    return_dict = {}
    for k, v in input_dict.items():
        quantity, name = title_to_quantity(k, return_name=True)
        if quantity.dimensionless == False:
            return_dict[name] = quantity * v
        else:
            return_dict[k] = v

    return return_dict

def cname_from_unit(data, unit):
    """
    Extracts the column name from a pandas array where the column name has a label indicating the desired unit.

    :param data: Pandas Dataframe containing the column you want to extract
    :param unit: pint unit you would like to match the column to (i.e. ureg.mV)
    """
    column_names = data.columns.values
    desired_unit_std = to_standard_quantity(1*unit).units

    column_quantities = [(i, title_to_quantity(c)) for i, c in enumerate(column_names)]
    column_units_std = [(i, to_standard_quantity(q).units) for (i, q) in column_quantities if q]
    column_matches = [(i, u == desired_unit_std) for i, u in column_units_std]
    column_indices = [i for i, b in column_matches if b == True]
    if len(column_matches) == 0:
        raise ValueError(f'ERROR: Could not find unit {unit} in data. Available column names are {column_names}')
    else:
        index = column_indices[0]
        name = column_names[index]
        return name

def column_from_unit(data, unit):
    """
    Extracts the pandas column corresponding to a particular unit quantities. If there are multiple columns, it gives back the first one. Assumes columns names are of the form "Quantity Name (unit)", i.e. Voltage (mV) or Photocurrent (nA).

    :param data: Pandas dataframe to extract column from
    :param unit: Pint unit of the desired column. Data will be returned in terms of this unit
    """
    column_name = cname_from_unit(data, unit)
    column_unit = title_to_quantity(column_name).to(unit)
    return column_unit * data[column_name].values



from numpy.testing import assert_equal, assert_allclose
from pandas.testing import assert_frame_equal
import numpy as np
import pandas as pd
from sciparse import find_lcr_dataline, parse_lcr_header, parse_lcr
from sciparse import convert_lcr_to_standard
import pytest
import os

@pytest.fixture
def filename():
    dir_name = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(dir_name, 'data/lcr_test_data.dat')
    filename = str(filename)
    return filename

@pytest.fixture
def metadata(filename):
    metadata = parse_lcr_header(filename)
    return metadata

@pytest.fixture
def data(filename):
    data, metadata = parse_lcr(filename)
    return data

def test_extract_header(metadata):
    desiredMode = "SWEEP"
    actualMode = metadata['mode']
    assert_equal(actualMode, desiredMode)

    desiredStartVoltage = 10
    actualStartVoltage = metadata['start_voltage']
    assert_equal(actualStartVoltage, desiredStartVoltage, err_msg="stop voltage")

    desiredStopVoltage = -20
    actualStopVoltage = metadata['stop_voltage']
    assert_equal(actualStopVoltage, desiredStopVoltage, err_msg="start voltage")

    desiredStepVoltage = -0.25
    actualStepVoltage = metadata['step_voltage']
    assert_equal(actualStepVoltage, desiredStepVoltage, err_msg="step voltage")

    desiredPoints = 121
    actualPoints = metadata['n_samples']
    assert_equal(actualPoints, desiredPoints, err_msg="number points")

def test_find_datalines(filename):
    desiredStartLine = 28
    actualStartLine = find_lcr_dataline(filename)
    assert_equal(actualStartLine, desiredStartLine)

def test_parse_data_header(data):
    # Confirm we got the right data types
    actualDataTypes = data.columns.values
    desiredDataTypes = ['Z', 'THETA', 'BIAS', 'VM', 'IM']
    assert_equal(actualDataTypes, desiredDataTypes)

def test_parse_data_length(data):
    # Confirm we got the right length of data
    desired_data_points = 121
    actual_data_points = len(data)
    assert_equal(actual_data_points, desired_data_points)

def test_parse_data(data):
    desiredZData = 5.57723*1e6
    actualZData = data['Z'].iloc[1]
    assert_allclose(desiredZData, actualZData)

    desiredBIASData = 8.5
    actualBIASData = data['BIAS'].iloc[6]
    assert_allclose(desiredBIASData, actualBIASData)

def test_convert_data_CP_RP():
    frequency = 1 / (2*np.pi) * 1000 # 1krad/s
    test_metadata = {'frequency': frequency}
    test_data = pd.DataFrame({'CP': [1e-9], 'RP': 1e6})
    desired_data = pd.DataFrame({
            'Z (ohm)': [1 / np.sqrt(2) * 1e6],
            'THETA (rad)': [-np.pi/4]})
    actual_data = convert_lcr_to_standard(test_data, test_metadata)
    assert_frame_equal(actual_data, desired_data)

def test_convert_data_CS_RS():
    frequency = 1 / (2*np.pi) * 1000 # 1krad/s
    test_metadata = {'frequency': frequency}
    test_data = pd.DataFrame({'CS': [1e-9], 'RS': 1e6})
    desired_data = pd.DataFrame({
            'Z (ohm)': [np.sqrt(2) * 1e6],
            'THETA (rad)': [-np.pi/4]})
    actual_data = convert_lcr_to_standard(test_data, test_metadata)
    assert_frame_equal(actual_data, desired_data)

def test_convert_data_C_Q():
    frequency = 1 / (2*np.pi) * 1000 # 1krad/s
    test_metadata = {'frequency': frequency}
    test_data = pd.DataFrame({'C': [1e-9], 'Q': 1})
    desired_data = pd.DataFrame({
            'Z (ohm)': [1 / np.sqrt(2) * 1e6],
            'THETA (rad)': [-np.pi/4]})
    actual_data = convert_lcr_to_standard(test_data, test_metadata)
    assert_frame_equal(actual_data, desired_data)

def test_convert_data_C_D():
    frequency = 1 / (2*np.pi) * 1000 # 1krad/s
    test_metadata = {'frequency': frequency}
    test_data = pd.DataFrame({'C': [1e-9], 'D': 1})
    desired_data = pd.DataFrame({
            'Z (ohm)': [1 / np.sqrt(2) * 1e6],
            'THETA (rad)': [-np.pi/4]})
    actual_data = convert_lcr_to_standard(test_data, test_metadata)
    assert_frame_equal(actual_data, desired_data)

def test_convert_raises_no_frequency():
    with pytest.raises(ValueError):
        test_data = pd.DataFrame({'C': [1e-9], 'D': 1})
        test_metadata = {}
        actual_data = convert_lcr_to_standard(test_data, test_metadata)

def test_convert_raises_wrong_model():
    with pytest.raises(ValueError):
        test_data = pd.DataFrame({'CCC': [1e-9], 'D': 1})
        test_metadata = {}
        actual_data = convert_lcr_to_standard(test_data, test_metadata)

